/* $Id: moorwin.c,v 1.1 1999/06/07 23:34:58 becka Exp $
******************************************************************************

   A simple game using LibGGI. 

   Copyright (C) 2000 Andreas Beck - becka@ggi-project.org
  
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>

#include <ggi/gg.h>
#include <ggi/ggi.h>
#include <ggi/ggi_ext.h>

#include "loadppm.h"

struct path {
	double x,y;
	double vx,vy;
	int currframe;
};

#define MAXLOGOS 10

typedef struct {
	int x,y;
} coord;

typedef struct {
	int numframes;
	sprite16 **frames;
} animsprite;

struct gamestate {

	coord	resolution;
	ggi_visual_t	vis;
	ggi_visual_t	memvis;
	
	double 		gunnerpos;
	double		fps;

	sprite16	*sight;
	coord		sight_hotspot;
	
	int numlayers;

	struct layers {
		sprite16	*layer;
		double		 faktor;
		int		 dy,h;

		animsprite	left,right,fall;
		struct path 	logo[MAXLOGOS];
	} *layers;

} gamestate;


int Sprite16Blit(sprite16 *source,int x,int y,int w,int h,
		ggi_visual_t target,int xt,int yt,int wt,int ht,
		int testx,int testy) {
	int xx,yy,rc;
	static const ggi_directbuffer   *dbuftarget;
	uint16 *rd,*wr,hlp;
	int rdskip,wrskip;

	rc=0;
	if (xt<0) { x-=xt;w+=xt;xt=0; }
	if (yt<0) { y-=yt;h+=yt;yt=0; }
	if (xt+w>wt) { w=wt-xt; }
	if (yt+h>ht) { h=ht-yt; }
	if (h<=0||w<=0) return 0;
	if (! (dbuftarget = ggiDBGetBuffer (target, 0))) 
		return -1;

        if (ggiResourceAcquire(dbuftarget->resource, GGI_ACTYPE_WRITE) != 0) 
        	return -1;

        rd=source->data;
        rd+=y*source->w+x;
        rdskip=source->w-w;

        wr=dbuftarget->write;
        wr+=(yt*dbuftarget->buffer.plb.stride)/2+xt;
        wrskip=dbuftarget->buffer.plb.stride/2-w;
        
	for(yy=0;yy<h;yy++) {
		for(xx=0;xx<w;xx++) {
			hlp=*rd++;
			if (hlp!=0xffff) {
				if (xt+xx==testx&&yt+yy==testy) rc=1;
				*wr++=hlp;
			} else	wr++;
		}
		rd+=rdskip;
		wr+=wrskip;
	}
	ggiResourceRelease(dbuftarget->resource);
	return rc;
}

char *fgets_nc(char *data, int size, FILE *stream) {
	char *rc;
	while (NULL!=(rc=fgets(data,size,stream))) {
		if (*data!='#'&&*data!='\n'&&*data!='\0') break;
	}
	return rc;
}

int load_general(FILE *config) {
	char data[1024],name[1024];
	int rc=0;
	while(!feof(config)) {
		if (NULL==fgets_nc(data,sizeof(data),config)) break;
		/*  */ if (strncmp(data,"endsection",10)==0) break;
		  else if (strncmp(data,"mode"  ,4)==0) {
		  	sscanf(data+4,"%d %d",
		  		&gamestate.resolution.x,
		  		&gamestate.resolution.y);
		} else if (strncmp(data,"layers",6)==0) {
		  	sscanf(data+6,"%d",
		  		&gamestate.numlayers);
		  	gamestate.layers=malloc(sizeof(gamestate.layers[0])*gamestate.numlayers);
		} else if (strncmp(data,"sight" ,5)==0) {
		  	sscanf(data+5," \"%[^\"]\"%d %d",
		  		name,
		  		&gamestate.sight_hotspot.x,
		  		&gamestate.sight_hotspot.y);
			gamestate.sight=loadsprite16(name);
		} else {
			printf("Unknown theme general directive: %s",data);rc=1;break;
		}
	}
	return rc;
}

int load_anim(char *names,animsprite *target) {
	int num,x;
	char *hlp, name[1024];

	num=0;hlp=names;
	while(*hlp) if (*hlp++=='"') num++;
	num/=2;

	target->numframes=num;
	num=0;
	target->frames=malloc(num*sizeof(target->frames[0]));
	
	while(1==sscanf(names," \"%[^\"]\"%n",name,&x)) {
		names+=x;
		if (NULL==(target->frames[num]=loadsprite16(name))) {
			printf("Cannot load animation component: %s\n",name);
			return 1;
		}
		num++;
	}
	return 0;
}

int currlayer=0;

int load_layer(FILE *config) {
	char data[1024],name[1024];
	int rc=0,j;
	for(j=0;j<MAXLOGOS;j++) {
		gamestate.layers[currlayer].logo[j].x=-1.0;
	}
	while(!feof(config)) {
		if (NULL==fgets_nc(data,sizeof(data),config)) break;
		/*  */ if (strncmp(data,"endsection",10)==0) break;
		  else if (strncmp(data,"layer" ,5)==0) {
		  	if (3!=sscanf(data+5," \"%[^\"]\"%d %d",
		  		name,
		  		&gamestate.layers[currlayer].dy,
		  		&gamestate.layers[currlayer].h)) {
				printf("Malformed layer directive: %s",data);
		  		rc=1;break;
		  	}
			gamestate.layers[currlayer].h-=gamestate.layers[currlayer].dy;
			if (NULL==(gamestate.layers[currlayer].layer=
				loadsprite16(name))) {
				printf("Cannot load layer: %s\n",name);
		  		rc=1;break;
			}
			gamestate.layers[currlayer].faktor=(gamestate.layers[currlayer].layer->w-gamestate.resolution.x);
		} else if (strncmp(data,"left" ,4)==0) {
			if (load_anim(data+4,&gamestate.layers[currlayer].left)) {
		  		rc=1;break;
			}
		} else if (strncmp(data,"right" ,5)==0) {
			if (load_anim(data+5,&gamestate.layers[currlayer].right)) {
		  		rc=1;break;
			}
		} else if (strncmp(data,"fall" ,4)==0) {
			if (load_anim(data+4,&gamestate.layers[currlayer].fall)) {
		  		rc=1;break;
			}
		} else {
			printf("Unknown theme layer directive: %s",data);rc=1;break;
		}
	}
	return rc;
}

int load_data(void) {
	FILE *config;
	char data[1024];
	int rc=0;

	currlayer=0;
	if (NULL==(config=fopen("theme","r"))) {
		fprintf(stderr,"Could not open theme file. Sorry.\n");
		return 1;
	}
	while(!feof(config)) {
		if (NULL==fgets_nc(data,sizeof(data),config)) break;
		/*  */ if (strncmp(data,"section general",15)==0) {
			if (load_general(config)) {rc=1;break;}
		} else if (strncmp(data,"section layer",13)==0) {
			if (load_layer(config)) {rc=1;break;}
			currlayer++;
		} else {
			printf("Unknown theme directive: %s",data);rc=1;break;
		}
	}
	fclose(config);
	gamestate.fps=25;
	return rc;
}

void setup_visual(void) {

	ggi_mode mode;

	/* Initialize LibGGI */
	if (ggiInit() != 0) {
		fprintf(stderr, "Unable to initialize LibGGI\n");
		exit(1);
	}
	
	/* Open the default visual */
	if ((gamestate.vis = ggiOpen(NULL)) == NULL) {
		ggiPanic("Unable to open default visual\n");
	}

	if ((gamestate.memvis = ggiOpen("display-memory")) == NULL) {
		ggiPanic("Unable to open memory backbuffer visual\n");
	}
	
	/* Turn on asynchronous mode (which should always be used) */
	ggiSetFlags(gamestate.vis, GGIFLAG_ASYNC);
	
	/* Set the default mode */
	if (ggiSetSimpleMode(gamestate.vis, gamestate.resolution.x, gamestate.resolution.y, 1, GT_16BIT) < 0) {
		ggiPanic("Unable to set %dx%dx16 mode\n",gamestate.resolution.x,gamestate.resolution.y);
	}

	if (ggiSetSimpleMode(gamestate.memvis, gamestate.resolution.x, gamestate.resolution.y, 1, GT_16BIT) < 0) {
		ggiPanic("Unable to set %dx%dx16 mode on backbuffer\n",gamestate.resolution.x,gamestate.resolution.y);
	}
	
	/* Check what mode we actually got */
	ggiGetMode(gamestate.vis, &mode);
}

void cleanup_visual(void) {

	/* Close visual */
	ggiClose(gamestate.memvis);
	ggiClose(gamestate.vis);

	/* Deinitialize LibGGI */
	ggiExit();
}

void update_gun(double gundelta) {
	if (gundelta<0.0) {
		if (gamestate.gunnerpos>-gundelta)
			gamestate.gunnerpos+=gundelta;
		else
			gamestate.gunnerpos =0.00;
	}
	if (gundelta>0.0) {
		if (gamestate.gunnerpos<1.0-gundelta)
			gamestate.gunnerpos+=gundelta;
		else
			gamestate.gunnerpos =1.00;
	}
}

int gameloop(void)
{
	ggi_event ev;
	int n,i,j,x,y;
	int hitlayer,hitlogo;
	double gundelta_key=0.0;
	double gundelta_mouse=0.0;
	int mousex,mousey;
	int shoot;
	struct timeval tod,new;

	mousex=gamestate.resolution.x/2;mousey=gamestate.resolution.y/2;

	gettimeofday(&tod,NULL);

	while(1) {
		shoot=0;
		tod.tv_usec+=1000000/gamestate.fps;
		if (tod.tv_usec>1000000) {
			tod.tv_usec-=1000000;tod.tv_sec++;
		}
		n=0;
		do {
			gettimeofday(&new,NULL);
			new.tv_sec =tod.tv_sec -new.tv_sec;
			new.tv_usec=tod.tv_usec-new.tv_usec;
			if (new.tv_usec<0) {
				new.tv_usec+=1000000;
				new.tv_sec--;
			}
			if (new.tv_sec<0) { 
				new.tv_sec=new.tv_usec=0;
				if (n==0) fprintf(stderr,"Cannot sustain framerate - down to %f.\n",gamestate.fps-=0.99);
				gettimeofday(&tod,NULL);
			}
			ggiEventPoll(gamestate.vis, emAll, &new);
			n++;
		} while(new.tv_usec);
		
		n = ggiEventsQueued(gamestate.vis, emAll);

		while (n) {
			ggiEventRead(gamestate.vis, &ev, emAll);
			
			switch(ev.any.type) {
				case evKeyPress:
				case evKeyRepeat:
					switch(ev.key.sym) {
						case 'q':
						case 'Q':
						case 27: return -1;
						case 'l': /* level cheat */
						case 'L': return 0;
						case GIIK_Left:
							gundelta_key=-0.02;
							break;
						case GIIK_Right:
							gundelta_key= 0.02;
							break;
					}
					break;
				case evKeyRelease:
					switch(ev.key.sym) {
						case GIIK_Left:
						case GIIK_Right:
							gundelta_key=0.0;
							break;
					}
					break;
				case evPtrRelative:
					mousex+=ev.pmove.x;
					mousey+=ev.pmove.y;
					if (mousex<0) mousex=0;
					else if (mousex>=gamestate.resolution.x) mousex=gamestate.resolution.x-1;
					if (mousey<0) mousey=0;
					else if (mousey>=gamestate.resolution.y) mousey=gamestate.resolution.y-1;
					break;
				case evPtrAbsolute:
					mousex=ev.pmove.x;
					mousey=ev.pmove.y;
					break;
				case evPtrButtonPress:
					if (ev.pbutton.button==GII_PBUTTON_PRIMARY)
						shoot=1;
					break;
			}
			n--;
		}

		hitlayer=hitlogo=-1;
		if (mousex<gamestate.resolution.x/10) gundelta_mouse=-0.02;
		else if (mousex>gamestate.resolution.x-gamestate.resolution.x/10) gundelta_mouse=+0.02;
		else gundelta_mouse=+0.0;

		update_gun(gundelta_key);
		update_gun(gundelta_mouse);

		for(i=0;i<gamestate.numlayers;i++) {
			struct layers *layer;
			
			layer=&gamestate.layers[i];
			
			if (1==Sprite16Blit(layer->layer,
				gamestate.gunnerpos*layer->faktor,0,
				gamestate.resolution.x,layer->h,
				gamestate.memvis,0,layer->dy,
				gamestate.resolution.x,gamestate.resolution.y,
				mousex,mousey
				)) {
				hitlayer=i;hitlogo=-1;
			}

			for (j=0;j<MAXLOGOS;j++) {
				animsprite *logo;
				struct path *path;

				path=&gamestate.layers[i].logo[j];

				if (path->x<0.0||path->x>layer->faktor+gamestate.resolution.x||path->y>gamestate.resolution.y||path->y<0.0) {
					switch(rand()%3) {
						case 0:
							path->x=0.0;
							path->y=rand()%gamestate.resolution.y;
							path->vx=(rand()%50+1)/10000.0*gamestate.layers[i].faktor;
							path->vy=0;
							break;
						case 1:
							path->x=layer->faktor+gamestate.resolution.x;
							path->y=rand()%gamestate.resolution.y;
							path->vx=-(rand()%50+1)/10000.0*gamestate.layers[i].faktor;
							path->vy=0;
							break;
						case 2:
							path->x=rand()%((int)layer->faktor+gamestate.resolution.x);
							path->y=gamestate.resolution.y;
							path->vx=(rand()%101-50)/10000.0*gamestate.layers[i].faktor;
							path->vy=-1;
							break;
					}
				}

				logo=(path->vx<0) ? &gamestate.layers[i].left
					 : (path->vx>0) ? &gamestate.layers[i].right
					 	: &gamestate.layers[i].fall ;

				path->currframe++;path->currframe%=logo->numframes;
				path->x+=path->vx;
				path->y+=path->vy;
				x=path->x-gamestate.gunnerpos*gamestate.layers[i].faktor;
				y=path->y;
				if (x>=gamestate.resolution.x||y>=gamestate.resolution.y) continue;
				if (1==Sprite16Blit(logo->frames[path->currframe],
					0,0,
					logo->frames[path->currframe]->w,
					logo->frames[path->currframe]->h,
					gamestate.memvis,x,y,
					gamestate.resolution.x,gamestate.resolution.y,
					mousex,mousey
					)) {
					hitlayer=i;hitlogo=j;
				}
			}
		}
		Sprite16Blit(gamestate.sight,
			0,0,
			gamestate.sight->w,gamestate.sight->h,
			gamestate.memvis,mousex-15,mousey-15,
			gamestate.resolution.x,gamestate.resolution.y,
			0,0
			);
		ggiCrossBlit(gamestate.memvis,0,0,
			gamestate.resolution.x,gamestate.resolution.y,
			gamestate.vis,0,0);
		if (shoot&&hitlayer>=0&&hitlogo>=0) {
			gamestate.layers[hitlayer].logo[hitlogo].vx=0.0;
			gamestate.layers[hitlayer].logo[hitlogo].vy=0.01*gamestate.layers[hitlayer].faktor;
		}
		ggiFlush(gamestate.vis);
	}
	return 1;
}


int main(void)
{
	int level;

	load_data();
	setup_visual();

	ggiSetGCForeground(gamestate.vis, 0);
	ggiFillscreen(gamestate.vis);
	switch(gameloop()) {
		case 0: /* Level completed */
			break;
		case 1: /* Lost ball */
			level--;break;
		default:	/* explicit quit */
			level=1000;break;
	}

	cleanup_visual();
	return 0;
}
